import torch
from typing import List, Any
from torch import Tensor

from elsciRL.encoders.encoder_abstract import StateEncoder

class StateEncoder():
    def __init__(self, possible_states):
        """Encoder for default state representation produced by the environment/engine."""
        self.all_possible_states = possible_states
        # Create dict lookup
        # - get binary list that indexes the state e.g. 0_0 -> [1,0,0,0] or 0_3 -> [0,0,0,1]
        self.enc = {}
        for i,s in enumerate(possible_states):
            self.enc[s] = [1 if i == j else 0 for j in range(len(possible_states))]

        device = "cuda" if torch.cuda.is_available() else "cpu" # Make this optional choice with parameter
        self.vectors: Tensor = torch.cat([torch.eye(len(self.all_possible_states)), torch.zeros(1,len(self.all_possible_states))]).to(device)         # tensor needs to be defined to len(local_object)
    
    def encode(self, state:Any = None, legal_actions:list = None, episode_action_history:list = None,
               indexed: bool = False) -> Tensor:
        """ NO CHANGE - Board itself is used as state as is and simply converted to a vector"""
        # Goes through every possible state and labels if occurance of state matches
        # Binary vector
        # NOT RECOMMENDED FOR LARGE STATE SPACES
        state_encoded = self.enc[state]
        state_encoded = torch.tensor(state_encoded)
        if (not indexed):
            state_encoded = self.vectors[state_encoded].flatten()

        return state_encoded    


import os
import pandas as pd
import numpy as np
import json
import random

# ------ Interaction Protocol -----------------------------------
from elsciRL.interaction_loops.standard import StandardInteractionLoop
# ------ Experiment Import --------------------------------------
from elsciRL.evaluation.standard_report import Evaluation
# ------ Agent Imports -----------------------------------------
# Universal Agents
from elsciRL.agents.agent_abstract import Agent, QLearningAgent
from elsciRL.agents.table_q_agent import TableQLearningAgent
from elsciRL.agents.neural_q_agent import NeuralQLearningAgent
# ------ Gym Experiement ----------------------------------------
from elsciRL.experiments.GymExperiment import GymExperiment
from elsciRL.agents.stable_baselines.DQN import SB_DQN
from elsciRL.agents.stable_baselines.PPO import SB_PPO
from elsciRL.agents.stable_baselines.A2C import SB_A2C

# TODO: Enable any number of the same agent types with varying parameters
AGENT_TYPES = {
    "Qlearntab": TableQLearningAgent,
    "DQN": NeuralQLearningAgent,
    "DQN_2": NeuralQLearningAgent,
    "DQN_language": NeuralQLearningAgent,
    "Random": random,
    "SB3_DQN": SB_DQN,
    "SB3_PPO": SB_PPO,
    "SB3_A2C": SB_A2C
}

PLAYER_PARAMS = {
    "Qlearntab": ["alpha", "gamma", "epsilon"],
    "DQN": ["input_type", "input_size", "sent_hidden_dim", "hidden_dim", "num_hidden", "sequence_size", "memory_size"],
    "DQN_2": ["input_type", "input_size", "sent_hidden_dim", "hidden_dim", "num_hidden", "sequence_size", "memory_size"],
    "DQN_language": ["input_type", "input_size", "sent_hidden_dim", "hidden_dim", "num_hidden", "sequence_size", "memory_size"],
    "Random": [],
    "SB3_DQN": ["policy"],
    "SB3_PPO": ["policy"],
    "SB3_A2C": ["policy"]
}

# This is the main run functions for elsciRL to be imported
# Defines the train/test operators and imports all the required agents and experiment functions ready to be used
# The local main.py file defines the [adapters, configs, environment] to be input

# This should be where the environment is initialized and then episode_loop (or train/test) is run
# -> results then passed down to experiment to produce visual reporting (staticmethod)
# -> instruction following approach then becomes alternative form of this file to be called instead
# -> This means we have multiple main.py types (e.g. with/without convergence measure) so should create a directory and finalise naming for this

class elsciRLOptimize:
    def __init__(self, Config:dict, LocalConfig:dict, 
                 Engine, Adapters:dict,
                 save_dir:str, show_figures:str, window_size:float,
                 instruction_path: dict=None, predicted_path: dict=None, instruction_episode_ratio:float=0.1,
                 instruction_chain:bool=False, instruction_chain_how:str='None'):
        self.ExperimentConfig = Config
        self.LocalConfig = LocalConfig

        

        try:
            self.setup_info = self.ExperimentConfig['data'] | self.LocalConfig['data']
        except:
            self.setup_info = self.ExperimentConfig | self.LocalConfig

        # If in experiment config make sure it pulls from this and not local config
        if 'adapter_input_dict' in self.ExperimentConfig:
            self.setup_info['adapter_input_dict'] = self.ExperimentConfig['adapter_input_dict']
        self.engine = Engine
        self.adapters = Adapters
        self.env = StandardInteractionLoop
        
        if not predicted_path:
            if not os.path.exists(save_dir):
                os.mkdir(save_dir)
            self.save_dir = save_dir+'/Instr_Experiment'
        else:
            save_dir_extra = save_dir.split("/")[-1]
            save_dir = '/'.join(save_dir.split("/")[:-1])
            if not os.path.exists(save_dir):
                os.mkdir(save_dir)
            self.save_dir = save_dir+'/'+save_dir_extra
        self.show_figures = show_figures
        if (self.show_figures.lower() != 'y')|(self.show_figures.lower() != 'yes'):
            print("Figures will not be shown and only saved.")  

        # New: Flag for if using gym agents to optimize instead
        # - Generates reward signal from instructions that is passed to gym eng translator
        # - init gym experiment if any gym agent selected
        self.is_gym_agent = {}
        for n,agent_type in enumerate(self.setup_info['agent_select']):
            if agent_type.split('_')[0] == "SB3":
                self.is_gym_agent[agent_type] = True
                self.sub_goal_reward = self.setup_info['reward_signal'][0]
                self.gym_exp = GymExperiment(Config=self.ExperimentConfig, ProblemConfig=self.LocalConfig, 
                        Engine=self.engine, Adapters=self.adapters,
                        save_dir=self.save_dir, show_figures = 'No', window_size=0.1)
                # Get start position to start instr chains
                train_setup_info = self.setup_info.copy()
                agent_type = "Qlearntab" # Force agent to Qlearntab for compatibility
                # Add Qlearntab if not existed and select first adapter
                if "Qlearntab" not in train_setup_info["adapter_input_dict"]:
                    first_agent_type = list(train_setup_info["adapter_input_dict"].keys())[0]
                    first_adapter = train_setup_info["adapter_input_dict"][first_agent_type][0]
                    train_setup_info["adapter_input_dict"]["Qlearntab"] = [first_adapter]
                adapter = train_setup_info["adapter_input_dict"][agent_type][0]
                # ----- Agent parameters
                agent_parameters = train_setup_info["agent_parameters"][agent_type]
                train_setup_info['agent_type'] = agent_type
                train_setup_info['agent_name'] = str(agent_type) + '_' + str(adapter) + '_' + str(agent_parameters)
                train_setup_info['adapter_select'] = adapter
                agent = AGENT_TYPES[agent_type](**agent_parameters)
                train_setup_info['agent'] = agent
                train_setup_info['train'] = True
                train_setup_info['live_env'] = True
                train_setup_info['training_results'] = False
                train_setup_info['observed_states'] = False
                train_setup_info['experience_sampling'] = False
                live_env = self.env(Engine=self.engine, Adapters=self.adapters, local_setup_info=train_setup_info)
                self.start_obs = live_env.start_obs
            else:
                self.is_gym_agent[agent_type] = False

        self.training_setups: dict = {}
        # New instruction learning
        # New for when we have predicted goal states for training
        if predicted_path:
            self.instruction_path = predicted_path
            if instruction_path:
                self.testing_path = instruction_path
            else:
                self.testing_path = predicted_path
        else:
            self.instruction_path = instruction_path
            self.testing_path = False   
            # New - instruction chaining
            # - i.e. when learning the env should not reset to original start position
            # - instead, next instruction starts from where previous ended
            self.instruction_chain = instruction_chain
            if instruction_chain_how == 'None':
                self.instruction_chain_how = 'Random'
            else:
                self.instruction_chain_how = instruction_chain_how

            # new - store agents cross training repeats for completing the same start-end goal
            self.trained_agents: dict = {}
            self.num_training_seeds = self.setup_info['number_training_seeds']
            # new - config input defines the re-use of trained agents for testing: 'best' or 'all'
            self.test_agent_type = self.setup_info['test_agent_type']
            self.analysis = Evaluation(window_size=window_size)
            # Defines the number of episodes used for sub-instructions
            self.instruction_episode_ratio = instruction_episode_ratio

        # Instruction Knowledge from Search
        self.known_instructions = list(self.instruction_path.keys())
        self.known_instructions_dict = {}
        # Extract sub-goal completion for each instruction based on search results
        agent_adapter_i = None
        for instr in self.known_instructions:
            start = instr.split("---")[0]
            end = instr.split("---")[1]
            for n, agent_type in enumerate(self.setup_info['agent_select']):
                print("SELECTED AGENTS",self.setup_info['agent_select'])
                print("ADAPTER LIST",self.setup_info["adapter_input_dict"])
                adapter_inputs = self.setup_info["adapter_input_dict"][agent_type]
                for adapter in adapter_inputs:
                    agent_adapter = agent_type+'_'+adapter
                    if agent_adapter not in self.known_instructions_dict:
                        self.known_instructions_dict[agent_adapter] = {}
                    
                    if agent_adapter in self.instruction_path[instr]:
                        count = self.instruction_path[instr][agent_adapter]['count']
                        if start not in self.known_instructions_dict[agent_adapter]:
                            self.known_instructions_dict[agent_adapter][start] = {}
                            if end not in self.known_instructions_dict[agent_adapter][start]:
                                self.known_instructions_dict[agent_adapter][start][end] = count
                    else:
                        # Supplement alternative search agent for this
                        # - we need it to match agent_adapter lookup for later calls so simply copies the search knowledge
                        # Search agent+adapter is now independent from optimization agent, 
                        #  - will default to match 
                        #  - but if optimization agent not seen in search then alternative must be used
                        agent_adapter_list = {}
                        i = 0
                        for item in self.instruction_path[instr]:
                            agent_adapter_list[str(i)] = item
                            i+=1
                        
                        if (i>1)&(agent_adapter_i is None):
                            print("\n Agent + Adapter not used in instruction search, please select the search agent:")
                            print(agent_adapter_list)
                            agent_adapter_i = input("\t - Select the id number of the default search agent+adapter you wish to use:    ")
                        elif (i>1)&(agent_adapter_i is not None):
                            agent_adapter_i = agent_adapter_i
                        else:
                            print("Only one agent used in instruction search, defaulting to this.")
                            print(agent_adapter_list)
                            agent_adapter_i = '0'
                        agent_adapter_copy = agent_adapter_list[agent_adapter_i]
                        # Copy knowledge of chosen search agent+adapter
                        # - Instruction path to define sub_goal list
                        self.instruction_path[instr][agent_adapter] = self.instruction_path[instr][agent_adapter_copy].copy()
                        # - Known instructions dict to define meta-MDP planner
                        count = self.instruction_path[instr][agent_adapter_copy]['count']
                        if start not in self.known_instructions_dict[agent_adapter]:
                            self.known_instructions_dict[agent_adapter][start] = {}
                            if end not in self.known_instructions_dict[agent_adapter][start]:
                                self.known_instructions_dict[agent_adapter][start][end] = count
                    
        print("-----")
        print("Known human instruction inputs. ")
        print(self.known_instructions_dict)
        print(" - ")
        self.total_num_instructions = 0
        for instr in self.instruction_path:
            print("\n \t - ", instr, " -> ", list(self.instruction_path[instr].keys()))
            self.total_num_instructions+=1
        
   
    def train(self):
        if not os.path.exists(self.save_dir):
            os.mkdir(self.save_dir)

        for n, agent_type in enumerate(self.setup_info['agent_select']):
            # Added gym based agents as selection
            is_gym_agent = self.is_gym_agent[agent_type]
            if is_gym_agent:
                env_start = self.start_obs 
                start = str(env_start).split(".")[0]
                goal = start + "---" + "GOAL"
                print("Long-term Goal: ", goal)
                # ---- 
                # New reward signal passed to engine to generate gym reward
                reward_signal = {}
                while True:
                    max_count = 0
                    # Go through path and extract a a reward signal for each sub-instruction
                    # If search cant use agent, it will default to Qlearntab
                    agent_adapter = (self.setup_info["agent_select"][n]+'_'+self.setup_info["adapter_input_dict"][self.setup_info["agent_select"][n]][0])
                    if agent_adapter not in self.known_instructions_dict:
                        agent_adapter = "Qlearntab"+"_"+self.setup_info["adapter_input_dict"]["Qlearntab"][0]
                    if start in self.known_instructions_dict[agent_adapter]:
                        for end in self.known_instructions_dict[agent_adapter][start]:
                            if self.known_instructions_dict[agent_adapter][start][end] > max_count:
                                max_count = self.known_instructions_dict[agent_adapter][start][end]
                                instr = start + "---" + end
                                print("Sub-instr: ", instr) 

                        sub_goal = self.instruction_path[instr][agent_adapter]['sub_goal']
                        # Get reward signal for each sub-goal
                        # - Sub-goals are lists of all matching env labels
                        for sg in sub_goal:
                            reward_signal[sg] = self.sub_goal_reward#*np.round(1/i, 4) # e.g. r=1 --> 1/2, 1/3, 1/4, ..
                        # ---
                        start = end
                        prior_instr = instr
                    else:
                        break     
                # Apply signal using dict:= {obs:reward, obs:reward, ...}
                print("GYM REWARD SIGNAL: ", reward_signal)
                self.gym_exp.reward_signal = reward_signal
                # --- GYM EXPERIMENT TRAINING
                train_setup_info = self.setup_info.copy()
                for adapter in train_setup_info["adapter_input_dict"][agent_type]:
                    self.gym_exp.setup_info['agent_select'] = [agent_type] 
                    self.training_setups[agent_adapter] = self.gym_exp.train() 
            else:
                # We are adding then overriding some inputs from general configs for experimental setups
                train_setup_info = self.setup_info.copy()
                # TODO: fix experience sampling
                if train_setup_info['experience_sample_batch_ratio']>0:
                    print("NOTE - Experience Sampling feature not currently implemented and will not be used")
                    train_setup_info['experience_sample_batch_ratio'] = 0
                # ----- State Adapter Choice
                for adapter in train_setup_info["adapter_input_dict"][agent_type]:
                    # ----- Agent parameters
                    agent_parameters = train_setup_info["agent_parameters"][agent_type]
                    train_setup_info['agent_type'] = agent_type
                    train_setup_info['agent_name'] = str(agent_type) + '_' + str(adapter) + '_' + str(agent_parameters)
                    train_setup_info['adapter_select'] = str(adapter)
                    print("+++++++++++++++++++")
                    print(agent_type,adapter)
                    # ----- Sub-Goal
                    # - If we have setup dict to include agent_adapter specific location of sub-goals
                    #   i.e. {instr:{env_code:{agent_adapter:{sub_goal:'ENV_CODE', sim_score:0.8}}, action_cap:5}}
                    #   Otherwise is standard user defined input {instr:{env_code:'ENV_CODE', action_cap:5}}
                    # -----
                    # Repeat training
                    train_setup_info['train'] = True
                    number_training_episodes = train_setup_info['number_training_episodes']
                    number_training_repeats = self.ExperimentConfig["number_training_repeats"]
                    print("Training Agent " + str(agent_type) + " for " + str(number_training_repeats) + " repeats")
                    if str(agent_type) + '_' + str(adapter) not in self.trained_agents:
                        self.trained_agents[str(agent_type) + '_' + str(adapter)] = {}

                    seed_recall = {}
                    seed_results_connection = {}
                    if self.num_training_seeds <1:
                        self.num_training_seeds = 1
                    for seed_num in range(0,self.num_training_seeds):
                        if self.num_training_seeds > 1:
                            print("------")
                            print("- Seed Num: ", seed_num)
                        # -------------------------------------------------------------------------------
                        # Initialise Environment
                        # Environment now init here and called directly in experimental setup loop
                        # - Observed states, experience sampling passed over seeds but not training repeats
                        if seed_num==0:
                            train_setup_info['training_results'] = False
                            train_setup_info['observed_states'] = False
                            train_setup_info['experience_sampling'] = False
                        else:
                            train_setup_info['training_results'] = False
                            train_setup_info['observed_states'] = observed_states_stored.copy()
                            train_setup_info['experience_sampling'] = experience_sampling_stored.copy()
                        # --- 
                        setup_num:int = 0
                        temp_agent_store = {}
                        for training_repeat in range(1,(number_training_repeats+1)):
                            if number_training_repeats > 1:
                                print("------")
                                print("- Repeat Num: ", training_repeat)
                            setup_num+=1
                            
                            # ----- init agent
                            player = AGENT_TYPES[agent_type](**agent_parameters)
                            train_setup_info['agent'] = player
                            
                            # init live environment
                            train_setup_info['live_env'] = True
                            live_env = self.env(Engine=self.engine, Adapters=self.adapters, local_setup_info=train_setup_info)
                            # ----------------
                            #  Start obs reset between seeds but fixed for repeat
                            if training_repeat > 1:
                                live_env.start_obs = env_start

                            env_start = live_env.start_obs
                            goal = str(env_start).split(".")[0] + "---" + "GOAL"
                            print("Long-term Goal: ", goal)
                            if goal in seed_recall:
                                setup_num = seed_recall[goal]
                            else:
                                seed_recall[goal] = 1
                            # Load results table from previous seeds to continue output graphs
                            if goal in seed_results_connection:
                                live_env.results.load(seed_results_connection[goal])
                        
                            # - Results save dir -> will override for same goal if seen in later seed
                            if self.num_training_seeds > 1:
                                agent_save_dir = self.save_dir+'/'+agent_type+'_'+adapter+'__training_results_'+str(goal)+'_'+str(setup_num) 
                            else:
                                agent_save_dir = self.save_dir+'/'+agent_type+'_'+adapter+'__training_results_'+str(setup_num)
                            if not os.path.exists(agent_save_dir):
                                os.mkdir(agent_save_dir)
                            
                            if train_setup_info['experience_sample_batch_ratio']>0:
                                live_sample_batch_size = int(number_training_episodes*train_setup_info['experience_sample_batch_ratio'])
                                live_sample_batch_count = int(1/train_setup_info['experience_sample_batch_ratio'])
                                # Train on Live system for limited number of total episodes   
                                live_env.num_train_episodes = live_sample_batch_size
                                print("-- Training with Simulated Batches, ", live_sample_batch_count, " total...")
                                # init simulated environment
                                train_setup_info['live_env'] = False
                                simulated_env = self.env(Engine=self.engine, Adapters=self.adapters, local_setup_info=train_setup_info)
                                simulated_env.start_obs = env_start

                                sample_batch_agent_store = {} # Need to continue learning across batches
                                sample_batch_results_store = {} # Pass over results table to next batch for the same instr TODO not working
                                for live_sample_batch in range(0, live_sample_batch_count-1):
                                    print("--- Live Interaction Batch Num", live_sample_batch+1)
                                    # ---- 
                                    # Train for sub instr plan
                                    start = str(env_start).split(".")[0]
                                    i=0
                                    while True:
                                        # Only allow insutrction up until total limi
                                        # - Prevents it being given more episodes than flat
                                        # - Prevents cyclic instruction paths
                                        if int(number_training_episodes*self.instruction_episode_ratio)<=(number_training_episodes-total_instr_episodes):
                                            # Override with trained agent if goal seen previously
                                            if goal in self.trained_agents[str(agent_type) + '_' + str(adapter)]:
                                                live_env.agent = self.trained_agents[str(agent_type) + '_' + str(adapter)][goal].clone()
                                                # Reset exploration parameter between seeds so to not get 'trapped'
                                                live_env.agent.exploration_parameter_reset()
                                            break
                                        
                                        elif start in self.known_instructions_dict[(agent_type+'_'+adapter)]:
                                            i+=1
                                            max_count = 0
                                            for end in self.known_instructions_dict[(agent_type+'_'+adapter)][start]:
                                                if self.known_instructions_dict[(agent_type+'_'+adapter)][start][end] > max_count:
                                                    max_count = self.known_instructions_dict[(agent_type+'_'+adapter)][start][end]
                                                    instr = start + "---" + end
                                                    print("Sub-instr: ", instr)
                                            # ---
                                            # Override trained agent with known instruction agent 
                                            if instr in sample_batch_agent_store:
                                                live_env.agent = sample_batch_agent_store[instr]
                                            elif instr in self.trained_agents[str(agent_type) + '_' + str(adapter)]:
                                                live_env.agent = self.trained_agents[str(agent_type) + '_' + str(adapter)][instr].clone()
                                            if instr in sample_batch_results_store:
                                                live_env.results.load(sample_batch_results_store[instr])
                                            
                                            # elif prior_instr:
                                            #     live_env.agent = live_env.agent.clone() # Start from agent from previous task
                                            # TODO: ADOPT AGENT OF MOST SIMILAR POLICY
                                            # ---
                                            sub_goal = self.instruction_path[instr][agent_type+'_'+adapter]['sub_goal']
                                            live_env.sub_goal = sub_goal
                                            live_env.agent.exploration_parameter_reset()
                                            training_results = live_env.episode_loop()
                                            sample_batch_agent_store[instr] = live_env.agent.clone()
                                            # Override episode numbers and store for next batch
                                            training_results['episode'] = training_results.index
                                            sample_batch_results_store[instr] = live_env.results.copy()
                                            # ---
                                            # Store instruction results
                                            instr_save_dir = agent_save_dir+'/'+str(i)+"-"+instr.replace(" ","").replace("/","_")
                                            if not os.path.exists(instr_save_dir):
                                                os.mkdir(instr_save_dir)

                                            # Produce training report with Analysis.py
                                            Return = self.analysis.train_report(training_results, instr_save_dir, self.show_figures)
                                            if instr not in temp_agent_store:
                                                temp_agent_store[instr] = {}
                                            temp_agent_store[instr][setup_num] = {'Return':Return,'agent':live_env.agent.clone()}
                                            # ---
                                            start = end
                                            live_env.results.reset() # Force reset so we don't get overlapping outputs
                                            #live_env.start_obs = sub_goal[0] # Set start position of env for next -> pick first sub-goal
                                        else:
                                            # if at least one sub-instr start with that for goal agent
                                            # if i > 1:
                                            #     live_env.agent = live_env.agent.clone() # Start from agent from previous task

                                            # if no instructions and first batch then adopt known agent
                                            # - Will then use this for next batches
                                            if i == 0 :
                                                if live_sample_batch == 0:
                                                    # Override with trained agent if goal seen previously
                                                    if goal in self.trained_agents[str(agent_type) + '_' + str(adapter)]:
                                                        live_env.agent = self.trained_agents[str(agent_type) + '_' + str(adapter)][goal].clone()
                                            break
                                    
                                    # train for entire path 
                                    #live_env.start_obs = env_start
                                    live_env.sub_goal = None 
                                    print("Goal: ", goal)
                                    if goal in sample_batch_results_store:
                                        live_env.results.load(sample_batch_results_store[goal])
                                    live_env.agent.exploration_parameter_reset()
                                    training_results = live_env.episode_loop()
                                    training_results['episode'] = training_results.index
                                    sample_batch_results_store[goal] = live_env.results.copy()

                                    # Train based on simulated exp             
                                    simulated_env.num_train_episodes = number_training_episodes
                                    # Connect live agent -> This should be linked so continues learning
                                    simulated_env.agent = live_env.agent
                                    simulated_env.episode_loop()
                                # Final batch doesn't require simulated exp after and we need result output
                                print("--- Final batch")
                                train_setup_info['live_env'] = True
                                train_setup_info['number_training_episodes'] = live_sample_batch_size
                                live_env.agent.exploration_parameter_reset()
                                training_results = live_env.episode_loop()

                                # Have to 'fix' output
                                training_results['episode'] = training_results.index
                                cumulative_r = 0
                                cumulative_r_lst = []
                                for r in training_results['episode_reward']:
                                    cumulative_r+=r
                                    cumulative_r_lst.append(cumulative_r)
                                training_results['cumulative_reward'] = cumulative_r_lst
                            else:
                                # ---- 
                                # Train for sub instr plan
                                # Set start position of env for next -> pick first sub-goal env label
                                start = str(env_start).split(".")[0]
                                i=0
                                total_instr_episodes = 0
                                instr_results = None
                                prior_instr = None
                                multi_sub_goal = {} # New multi-goal option -> needs to be defined in env
                                while True:
                                    i+=1
                                    if i > self.total_num_instructions:
                                        break
                                    max_count = 0
                                    # Only allow instruction up until total limit
                                    # - Prevents it being given more episodes than flat
                                    # - Prevents cyclic instruction paths
                                    if int(number_training_episodes*self.instruction_episode_ratio)>=(number_training_episodes-total_instr_episodes):
                                        # Override with trained agent if goal seen previously
                                        if goal in self.trained_agents[str(agent_type) + '_' + str(adapter)]:
                                            live_env.agent = self.trained_agents[str(agent_type) + '_' + str(adapter)][goal].clone()
                                            # Reset exploration parameter between seeds so to not get 'trapped'
                                            live_env.agent.exploration_parameter_reset()
                                        break
                                    
                                    elif start in self.known_instructions_dict[(agent_type+'_'+adapter)]:
                                        for end in self.known_instructions_dict[(agent_type+'_'+adapter)][start]:
                                            if self.known_instructions_dict[(agent_type+'_'+adapter)][start][end] > max_count:
                                                max_count = self.known_instructions_dict[(agent_type+'_'+adapter)][start][end]
                                                instr = start + "---" + end
                                                print("Sub-instr: ", instr)
                                                
                                        # Instructions use fewer episodes, lower bound to 10
                                        number_instr_episodes = int(number_training_episodes*self.instruction_episode_ratio)
                                        if number_instr_episodes<10:
                                            number_instr_episodes=10
                                        total_instr_episodes+=number_instr_episodes
                                        live_env.num_train_episodes = number_instr_episodes
                                        # ---
                                        # Override trained agent with known instruction agent
                                        if instr in self.trained_agents[str(agent_type) + '_' + str(adapter)]:
                                            live_env.agent = self.trained_agents[str(agent_type) + '_' + str(adapter)][instr].clone()
                                        # TODO: ADOPT AGENT OF MOST SIMILAR POLICY
                                        # ---
                                        # New: Allow env to start from prior instr end
                                        if self.instruction_chain:
                                            if prior_instr:
                                                # Select first env position from known sub-goal list
                                                if self.instruction_chain_how.lower() == 'first':
                                                    env_sg_start = self.instruction_path[prior_instr][agent_type+'_'+adapter]['sub_goal'][0]
                                                elif self.instruction_chain_how.lower() == 'random':
                                                    env_sg_start = random.choice(self.instruction_path[prior_instr][agent_type+'_'+adapter]['sub_goal'])
                                                elif self.instruction_chain_how.lower() == 'exact':
                                                    try:
                                                        if live_env.sub_goal_end:
                                                            env_sg_start = live_env.sub_goal_end
                                                        else:
                                                            env_sg_start = random.choice(self.instruction_path[prior_instr][agent_type+'_'+adapter]['sub_goal'])
                                                    except:
                                                        print("ERROR: To use the EXACT instruction chain, the environment must include the '.sub_goal_end' attribute.")
                                                elif (self.instruction_chain_how.lower() == 'continuous')|(self.instruction_chain_how.lower() == 'cont'):
                                                    # Env start position is default (i.e. reset)
                                                    env_sg_start = None
                                                    # Define multi-sub-goals
                                                    # - Scale prior instruction down based on current path by r = 1/x
                                                    # - Reward override if same instr seen later to prevent cyclic loops that don't complete episode 
                                                    if prior_instr not in multi_sub_goal:
                                                        multi_sub_goal[prior_instr] = {}
                                                        multi_sub_goal[prior_instr]['sub_goal'] = self.instruction_path[instr][agent_type+'_'+adapter]['sub_goal']
                                                    multi_sub_goal[prior_instr]['reward_scale'] = np.round(1/i, 4) # e.g. r=1 --> 1/2, 1/3, 1/4, ...
                                                        
                                                    try:
                                                        # This doesn't supercede .sub_goal so need both defined
                                                        live_env.multi_sub_goal = multi_sub_goal
                                                    except:
                                                        print("ERROR: To use CONTINUOUS instruction chain, the environment must include the '.multi_sub_gial' attribute.")
                                                
                                                if env_sg_start:
                                                    live_env.start_obs = env_sg_start
                                            
                                        sub_goal = self.instruction_path[instr][agent_type+'_'+adapter]['sub_goal']
                                        live_env.sub_goal = sub_goal
                                        live_env.agent.exploration_parameter_reset()
                                        if type(instr_results)==type(pd.DataFrame()):
                                            live_env.results.load(instr_results)
                                        training_results = live_env.episode_loop()
                                        training_results['episode'] = training_results.index
                                        # ---
                                        # Store instruction results
                                        instr_save_dir = agent_save_dir+'/'+str(i)+"-"+instr.replace(" ","").replace("/","_")
                                        if not os.path.exists(instr_save_dir):
                                            os.mkdir(instr_save_dir)

                                        # Produce training report with Analysis.py
                                        Return = self.analysis.train_report(training_results, instr_save_dir, self.show_figures)
                                        if instr not in temp_agent_store:
                                            temp_agent_store[instr] = {}
                                        temp_agent_store[instr][setup_num] = {'Return':Return,'agent':live_env.agent.clone()}
                                        # ---
                                        start = end
                                        prior_instr = instr
                                        # New: Dont reset results for each sub-instr so we show the training results with this included
                                        #live_env.results.reset() # Force reset so we don't get overlapping outputs
                                        instr_results =  live_env.results.copy()
                                    else:
                                        # If no instructions then train for full goal
                                        if i == 1:
                                            # Override with trained agent if goal seen previously
                                            if goal in self.trained_agents[str(agent_type) + '_' + str(adapter)]:
                                                live_env.agent = self.trained_agents[str(agent_type) + '_' + str(adapter)][goal].clone()
                                                # Reset exploration parameter between seeds so to not get 'trapped'
                                                live_env.agent.exploration_parameter_reset()
                                        break
                                # train for entire path 
                                if self.instruction_chain:
                                    live_env.start_obs = env_start
                                # Number of episodes used reduced by those used for instructions (lower bounded)
                                if (number_training_episodes-total_instr_episodes)<int(number_training_episodes*self.instruction_episode_ratio):
                                    if int(number_training_episodes*self.instruction_episode_ratio) < 10:
                                        live_env.num_train_episodes = 10
                                    else:
                                        live_env.num_train_episodes = int(number_training_episodes*self.instruction_episode_ratio)
                                else:
                                    live_env.num_train_episodes = number_training_episodes - total_instr_episodes
                                # Remove sub-goal
                                live_env.sub_goal = None
                                print("Goal: ", goal)
                                # Add instruction training to output chart
                                if type(instr_results)==type(pd.DataFrame()):
                                    live_env.results.load(instr_results)
                                
                                training_results = live_env.episode_loop()
                                training_results['episode'] = training_results.index
                                
                            # Opponent now defined in local setup.py
                            # ----- Log training setup      
                            training_results.insert(loc=0, column='Repeat', value=setup_num)                    
                            # Produce training report with Analysis.py
                            Return = self.analysis.train_report(training_results, agent_save_dir, self.show_figures)
                            if goal not in temp_agent_store:
                                temp_agent_store[goal] = {}
                            temp_agent_store[goal][setup_num] = {'Return':Return,'agent':live_env.agent.clone()}
                                                                    
                            # Extract trained agent from env and stored for re-call
                            # - Observed states and experience sampling from best repeat stored for next seed
                            if training_repeat == 1:
                                max_Return = Return
                                training_results_stored =  live_env.results.copy()
                                observed_states_stored = live_env.elsciRL.observed_states
                                experience_sampling_stored = live_env.elsciRL.experience_sampling
                            if Return > max_Return:
                                max_Return = Return
                                training_results_stored =  live_env.results.copy()
                                observed_states_stored = live_env.elsciRL.observed_states
                                experience_sampling_stored = live_env.elsciRL.experience_sampling
                            
                            seed_recall[goal] = seed_recall[goal] + 1
                        seed_results_connection[goal] = training_results_stored

                        for instr in temp_agent_store:
                            start_repeat_num = list(temp_agent_store[instr].keys())[0]
                            end_repeat_num = list(temp_agent_store[instr].keys())[-1]

                            if self.test_agent_type.lower() == 'best':
                                # Only save the best agent from repeated training
                                print("---------")
                                print("Selecting best agent from training repeats.")
                                best_return = temp_agent_store[instr][start_repeat_num]['Return']
                                best_agent = temp_agent_store[instr][start_repeat_num]['agent']
                                for repeat in range(start_repeat_num+1,end_repeat_num+1):
                                    if temp_agent_store[instr][repeat]['Return']>best_return:
                                        best_return = temp_agent_store[instr][repeat]['Return']
                                        best_agent = temp_agent_store[instr][repeat]['agent']
                                        print(best_return)
                        
                                if instr not in self.trained_agents[str(agent_type) + '_' + str(adapter)]:
                                    self.trained_agents[str(agent_type) + '_' + str(adapter)][instr] = {}
                                self.trained_agents[str(agent_type) + '_' + str(adapter)][instr] = best_agent      
                            elif self.test_agent_type.lower() == 'all':
                                all_agents = []
                                for repeat in range(start_repeat_num,end_repeat_num+1):
                                    agent = temp_agent_store[instr][repeat]['agent']
                                    all_agents.append(agent)
                                    
                                if instr not in self.trained_agents[str(agent_type) + '_' + str(adapter)]:
                                    self.trained_agents[str(agent_type) + '_' + str(adapter)][instr] = {}
                                self.trained_agents[str(agent_type) + '_' + str(adapter)][instr] = all_agents

                                                
                    # Store last train_setup_info as collection of observed states and experience sampling
                    self.training_setups['Training_Setup_'+str(agent_type) + '_' + str(adapter)] = train_setup_info.copy()
                #if (number_training_repeats>1)|(self.num_training_seeds):
                self.analysis.training_variance_report(self.save_dir, self.show_figures)
                    
        #json.dump(self.training_setups) # TODO: Won't currently serialize this output to a json file
        return self.training_setups

    # TESTING PLAY
    def test(self, training_setups:str=None):
        # Override input training setups with previously saved 
        if training_setups is None:
            training_setups = self.training_setups
            print("=== TESTING SETUPS ===")
            print(training_setups.keys())
        else:
            json.load(training_setups)

        for training_key in list(training_setups.keys()):   
            test_setup_info = training_setups[training_key]
            test_setup_info['train'] = False # Testing Phase
            agent_type = test_setup_info['agent_type']
            print("----------")
            print(training_key) 
            print("Testing results for trained agents in saved setup configuration:")
            print("TESTING SETUP INFO")
            print(test_setup_info['agent_type'])
            print(test_setup_info['adapter_select'])
            print("----------")
            agent_adapter = agent_type + "_" + test_setup_info['adapter_select']
            print(agent_adapter)
            # Added gym based agents as selection
            # - Strip reward signal from instructions for testing
            if self.is_gym_agent[agent_type]:
                gym_test_exp = self.training_setups[agent_adapter]
                gym_test_exp.reward_signal = None
                gym_test_exp.test()                        
            else:
                # Only use the trained agent with best return
                if self.test_agent_type.lower()=='best':
                    for testing_repeat in range(0, test_setup_info['number_test_repeats']):  
                        # Re-init env for testing
                        env = self.env(Engine=self.engine, Adapters=self.adapters, local_setup_info=test_setup_info)
                        # ---
                        start_obs = env.start_obs
                        goal = str(start_obs).split(".")[0] + "---" + "GOAL"
                        print("Flat agent Goal: ", goal)
                        # Override with trained agent if goal seen previously
                        if goal in self.trained_agents[test_setup_info['agent_type']+ '_' +test_setup_info['adapter_select']]:
                            print("Trained agent available for testing.")
                            env.agent = self.trained_agents[test_setup_info['agent_type']+'_'+test_setup_info['adapter_select']][goal]
                        else:
                            print("NO agent available for testing position.")
                        env.agent.epsilon = 0 # Remove random actions
                        # ---
                        # Testing generally is the agents replaying on the testing ENV
                        testing_results = env.episode_loop() 
                        test_save_dir = (self.save_dir+'/'+agent_adapter+'__testing_results_'+str(goal).split("/")[0]+"_"+str(testing_repeat))
                        print(test_save_dir)
                        if not os.path.exists(test_save_dir):
                            os.mkdir(test_save_dir)
                        # Produce training report with Analysis.py
                        Return = self.analysis.test_report(testing_results, test_save_dir, self.show_figures)
                        
                # Re-apply all trained agents with fixed policy
                elif self.test_agent_type.lower()=='all':
                    # All trained agents are used:
                    # - Repeats can be used to vary start position
                    # - But assumed environment is deterministic otherwise
                    # Re-init env for testing
                    for testing_repeat in range(0, test_setup_info['number_test_repeats']):
                        env = self.env(Engine=self.engine, Adapters=self.adapters, local_setup_info=test_setup_info)
                        # ---
                        start_obs = env.start_obs
                        goal = str(start_obs).split(".")[0] + "---" + "GOAL"
                        print("Flat agent Goal: ", goal)
                        # Override with trained agent if goal seen previously
                        if goal in self.trained_agents[test_setup_info['agent_type']+ '_' +test_setup_info['adapter_select']]:
                            print("Trained agents available for testing.")
                            all_agents = self.trained_agents[test_setup_info['agent_type']+'_'+test_setup_info['adapter_select']][goal]
                        else:
                            print("NO agent available for testing position.")
                        
                        for ag,agent in enumerate(all_agents):
                            env.results.reset() # Reset results table for each agent
                            env.start_obs = start_obs
                            env.agent = agent
                            env.agent.epsilon = 0 # Remove random actions
                            agent_adapter = test_setup_info['agent_type'] + "_" + test_setup_info['adapter_select']
                            # ---
                            # Testing generally is the agents replaying on the testing ENV
                            testing_results = env.episode_loop() 
                            test_save_dir = (self.save_dir+'/'+agent_adapter+'__testing_results_'+str(goal).split("/")[0]+"_"+"agent-"+str(ag)+"-repeat-"+str(testing_repeat))
                            if not os.path.exists(test_save_dir):
                                os.mkdir(test_save_dir)
                            # Produce training report with Analysis.py
                            Return = self.analysis.test_report(testing_results, test_save_dir, self.show_figures)

            # Path is the experiment save dir + the final instruction
            #if (number_training_repeats>1)|(self.test_agent_type.lower()=='all'):
            self.analysis.testing_variance_report(self.save_dir, self.show_figures)


    def render_results(self, training_setups:str=None):
        """Apply fixed policy to render current decision making for limited number of episodes."""
        # Override input training setups with previously saved 
        if training_setups is None:
            training_setups = self.training_setups
        else:
            json.load(training_setups)

        for training_key in list(training_setups.keys()):    
            test_setup_info = training_setups[training_key]
            test_setup_info['train'] = False # Testing Phase
            print("----------")
            print("Rendering trained agent's policy:")
            agent_adapter = test_setup_info['agent_type'] + "_" + test_setup_info['adapter_select']


            # Only use the trained agent with best return
            if self.test_agent_type.lower()=='best':
                # Re-init env for testing
                env = self.env(Engine=self.engine, Adapters=self.adapters, local_setup_info=test_setup_info)
                # ---
                start_obs = env.start_obs
                goal = str(start_obs).split(".")[0] + "---" + "GOAL"
                print("Flat agent Goal: ", goal)
                # Override with trained agent if goal seen previously
                if goal in self.trained_agents[test_setup_info['agent_type']+ '_' +test_setup_info['adapter_select']]:
                    print("Trained agent available for testing.")
                    env.agent = self.trained_agents[test_setup_info['agent_type']+'_'+test_setup_info['adapter_select']][goal]
                else:
                    print("NO agent available for testing position.")
                env.agent.epsilon = 0 # Remove random actions
                # ---
                # Testing generally is the agents replaying on the testing ENV
                render_save_dir = self.save_dir+'/render_results'
                if not os.path.exists(render_save_dir):
                    os.mkdir(render_save_dir)
                render_results = env.episode_loop(render=True, render_save_dir=render_save_dir)
                # Produce training report with Analysis.py
                #Return = self.analysis.test_report(render_results, render_save_dir, self.show_figures)
                
            # Re-apply all trained agents with fixed policy
            elif self.test_agent_type.lower()=='all':
                # All trained agents are used:
                # - Repeats can be used to vary start position
                # - But assumed environment is deterministic otherwise
                # Re-init env for testing
                env = self.env(Engine=self.engine, Adapters=self.adapters, local_setup_info=test_setup_info)
                # ---
                start_obs = env.start_obs
                goal = str(start_obs).split(".")[0] + "---" + "GOAL"
                print("Flat agent Goal: ", goal)
                # Override with trained agent if goal seen previously
                if goal in self.trained_agents[test_setup_info['agent_type']+ '_' +test_setup_info['adapter_select']]:
                    print("Trained agents available for testing.")
                    all_agents = self.trained_agents[test_setup_info['agent_type']+'_'+test_setup_info['adapter_select']][goal]
                else:
                    print("NO agent available for testing position.")
                
                for ag,agent in enumerate(all_agents):
                    env.results.reset() # Reset results table for each agent
                    env.start_obs = start_obs
                    env.agent = agent
                    # only run once
                    env.num_train_repeat = 1
                    env.num_test_repeat = 1
                    env.agent.epsilon = 0 # Remove random actions
                    agent_adapter = test_setup_info['agent_type'] + "_" + test_setup_info['adapter_select']
                    # ---
                    # Testing generally is the agents replaying on the testing ENV
                    render_save_dir = self.save_dir+'/render_results'
                    if not os.path.exists(render_save_dir):
                        os.mkdir(render_save_dir)
                    render_results = env.episode_loop(render=True, render_save_dir=render_save_dir)
                    # Produce training report with Analysis.py
                    #Return = self.analysis.test_report(render_results, render_save_dir, self.show_figures)




# from distutils.core import setup
from setuptools import setup, find_packages

setup(
    name='elsciRL',
    version='0.3.1',
    packages=[
        'elsciRL', 
        'elsciRL.adapters',
        'elsciRL.agents',
        'elsciRL.agents.stable_baselines',
        'elsciRL.analysis',
        'elsciRL.encoders', 
        'elsciRL.environment_setup',
        'elsciRL.evaluation',
        'elsciRL.examples',
        'elsciRL.examples.adapters',
        'elsciRL.examples.environments',
        'elsciRL.examples.local_configs',
        'elsciRL.GUI',
        'elsciRL.GUI.static',
        'elsciRL.GUI.templates',
        'elsciRL.GUI.LLM',
        'elsciRL.experiments',
        'elsciRL.instruction_following',
        'elsciRL.interaction_loops',
        'elsciRL.application_suite',
        ],
    package_data={
        'elsciRL.GUI.templates': ['index.html'],
        'elsciRL.GUI.static': ['styles.css', 'app_setup.md'],
    },
    include_package_data=True,
    url='https://github.com/pdfosborne/elsciRL',
    license='Apache-2.0 license',
    author='Philip Osborne',
    author_email='pdfosborne@gmail.com',
    description='Applying the elsciRL architecture to Reinforcement Learning problems.',
    install_requires=[
        'numpy',
        'pandas',
        'matplotlib',
        'seaborn',
        'scipy>=1.10.1',
        'torch',
        'tqdm',
        'httpimport',
        'sentence-transformers',
        'gymnasium',
        'stable-baselines3',
        'flask'
    ] 
)



import sys
import os
import shutil
from datetime import datetime
import json

# App tools
from flask import Flask, render_template, request, jsonify, send_from_directory

# elsci methods
from elsciRL.GUI.elsciRL_demo_search import elsciRLSearch as elsci_search
from elsciRL.instruction_following.elsciRL_instruction_following import elsciRLOptimize
from elsciRL.experiments.standard import Experiment as STANDARD_RL
# Get application data
from elsciRL.application_suite.import_data import Applications
from elsciRL.application_suite.import_tool import PullApplications
# Analysis
import matplotlib
matplotlib.use('Agg')
from elsciRL.analysis.combined_variance_visual import combined_variance_analysis_graph as COMBINED_VARIANCE_ANALYSIS_GRAPH

# LLM API Setup
from elsciRL.GUI.LLM.LLM_utils import call_gpt_api, process_gpt_response


dir_path = os.path.dirname(os.path.realpath(__file__))
app = Flask(__name__, static_folder=os.path.join(dir_path, 'static'), 
            template_folder=os.path.join(dir_path, 'templates'))


class WebApp:
    def __init__(self, save_dir: str = '', num_explor_epi: int = 1000):
        self.global_save_dir = save_dir
        self.num_explor_epi = num_explor_epi
        imports = Applications().data
        possible_applications = list(imports.keys())
        self.available_applications = possible_applications
        

        # Currently pulls all the available applications
        # - TODO: Make it so it pulls all the file names but not the data
        #    ---> Then once problem is selected it then pulls data
        self.application_data = PullApplications()
        self.pull_app_data = self.application_data.pull(problem_selection=possible_applications)
        self.config = self.application_data.setup()

        # Data used for LLM prompt
        with open(os.path.join(app.static_folder, 'app_setup.md'), "r") as f:
            self.app_setup_info = f.read()

    def load_data(self):
        # Init data here so it reset when page is reloaded
        self.global_input_count = 0
        self.instruction_results = {}
        self.instruction_results_validated = {}
        self.correct_instructions = []
        self.incorrect_instructions = []

        if not os.path.exists('./elsciRL-App-output'):
            os.mkdir('./elsciRL-App-output')
        if 'results' not in self.global_save_dir:
            time = datetime.now().strftime("%d-%m-%Y_%H-%M")
            save_dir = './elsciRL-App-output/' + str('results') + '_' + time
            if not os.path.exists(save_dir):                
                os.mkdir(save_dir)
            self.global_save_dir = save_dir

        if not os.path.exists(self.global_save_dir+'/uploads'):                
            os.mkdir(self.global_save_dir+'/uploads')
        print("GLOBAL SAVE DIR: ", self.global_save_dir)
        self.uploads_dir = os.path.abspath(os.path.join(self.global_save_dir, 'uploads'))
        if not os.path.exists(self.uploads_dir):
            os.makedirs(self.uploads_dir, exist_ok=True)
        print(f"Uploads directory (absolute path): {self.uploads_dir}")

    def home(self):
        template_path = os.path.join(app.template_folder, 'index.html')
        print(f"Trying to get HTML file from: {template_path}")
        return render_template('index.html')

    def get_applications(self):
        return jsonify({
            'applications': self.available_applications
        })
    
    def get_adapters(self, selected_application: str = ''):
        if selected_application == '':
            return []
        
        try:
            adapters = list(self.pull_app_data[selected_application]['adapters'].keys())
            return adapters
        except:
            print("Error fetching adapters...")
            return []
    
    def get_observed_states(self, selected_application):
        observed_states = list(self.pull_app_data[selected_application[0]]['prerender_data'].keys())
        return observed_states

    def get_local_configs(self, selected_application:str=''):
        if selected_application == '':
            return []
            
        try:
            local_configs = list(self.pull_app_data[selected_application]['local_configs'].keys())
            return local_configs
        except:
            print("Application data not found...")
            return []

    def get_plot_options(self, selected_application: str = ''):
        if selected_application == '':
            return []
        
        try:
            plot_options = list(self.pull_app_data[selected_application]['local_analysis'].keys())
            return plot_options
        except:
            print("Error fetching plot options...")
            return []

    def get_all_options(self):
        all_local_configs = {}
        all_observed_states = {}
        all_plot_options = {}
        all_experiment_configs = {}
        for app in self.available_applications:
            all_observed_states[app] = self.get_observed_states([app])
            all_local_configs[app] = self.get_local_configs(app)
            all_plot_options[app] = self.get_plot_options(app)
            try:
                all_experiment_configs[app] = list(self.pull_app_data[app]['experiment_configs'].keys())
            except Exception as e:
                print(f"Error fetching experiment configs for {app}: {e}")
                all_experiment_configs[app] = None
        return jsonify({
            'localConfigs': all_local_configs,
            'observedStates': all_observed_states,
            'plotOptions': all_plot_options,
            'experimentConfigs': all_experiment_configs
        })
    
    def generate_application(self, user_input:str=''):
        # TODO: Use this in a new tab with user input to update application list
        # Load the app_setup.md content as part of the system prompt
        
        # Add requirement to system prompt for code chunk separation
        system_prompt_requirement = (
            "If your response contains any code chunks, you must output them in a separate section clearly marked as 'Code Output', "
            "so that the application can extract and save them to a file. Do not mix code with explanations in the same section."
        )
        # Combine the app_setup.md info with the system prompt and the new requirement
        system_prompt = (
            "You are a helpful assistant. "
            "Below is important application setup information for elsciRL:\n"
            f"{self.app_setup_info}\n"
            f"{system_prompt_requirement}\n"
            "Please use this information to answer user queries."
        )
        
        if not user_input:
            return jsonify({"error": "No input provided"}), 400

        # Use the utils function to call the GPT API
        response = call_gpt_api(system_prompt + "\nUser: " + user_input)
        reply = process_gpt_response(response)
        print(reply)
        if not reply:
            return jsonify({"error": "Failed to get response from GPT API"}), 500

        # Save the complete output to a .txt file
        output_dir = os.path.join(os.path.dirname(__file__), 'output')
        os.makedirs(output_dir, exist_ok=True)
        output_path = os.path.join(output_dir, 'last_gpt_response.txt')
        with open(output_path, 'w', encoding='utf-8') as f:
            f.write(str(response))

        # Follow-up: Ask the AI model to extract all Python code and JSON config blocks and return a list of (filename, code) pairs
        followup_prompt = (
            "Extract all Python code blocks and JSON config blocks from the following text. "
            "For each code or config block, output a JSON array where each item has 'filename' and 'code' fields. "
            "Choose a descriptive filename for each code block (e.g., based on class/function names or comments, use .py for Python and .json for configs). "
            "Do not include any explanation, only the JSON array.\n\n" + reply
        )
        code_response = call_gpt_api(followup_prompt)
        code_reply = process_gpt_response(code_response)
        try:
            code_blocks = json.loads(code_reply)
            generated_data = {}
            for block in code_blocks:
                fname = block.get('filename', 'extracted_code.py')
                code = block.get('code', '')
                generated_data[fname] = code
                code_file_path = os.path.join(output_dir, fname)
                with open(code_file_path, 'w', encoding='utf-8') as code_file:
                    code_file.write(code)
        except Exception as e:
            # fallback: save the raw reply if not valid JSON
            code_file_path = os.path.join(output_dir, 'extracted_code.py')
            with open(code_file_path, 'w', encoding='utf-8') as code_file:
                code_file.write(code_reply.strip())

        for name,code in generated_data.items():
            if 'engine' in name.lower():
                generated_data['engine'] = code
            elif 'analysis' in name.lower():
                generated_data['analysis'] = code
            elif ('experiment' in name.lower()) | ('agent' in name.lower()):
                generated_data['agent_config'] = code
            elif ('local' in name.lower()) | ('env' in name.lower()):
                generated_data['local_config'] = code
            elif 'adapter_language' in name.lower():
                generated_data['adapter_language'] = code
            elif ('numeric' in name.lower()) | ('default' in name.lower()):
                generated_data['adapter_numeric'] = code
        
        # Create the application setup dictionary
        application_setup = {
            'engine':generated_data['engine'],
            'experiment_configs':{'quick_test':generated_data['agent_config']},
            'local_configs':{'env_config':generated_data['local_config']},
            'adapters':{'numeric_adapter':generated_data['adapter_numeric'],
                        'language_adapter':generated_data['adapter_language']},
            'local_analysis':{'blackjack_graphs':generated_data['analysis']},
            'prerender_data':{},
            'prerender_images':{},
        }

        # Add the new application to the application data
        self.pull_app_data = self.application_data.add_applicaiton(
            problem=generated_data['agent_config']['name'], 
            application_data=application_setup
            )

        return jsonify({"reply": reply})

    def process_input(self):
        data = request.get_json()
        if data is None:
            return jsonify({'error': 'No data provided'}), 400

        user_input = data.get('userInput', '')
        application = data.get('selectedApps', [])[0]
        config_input = data.get('localConfigInput', '')
        observed_states_filename = data.get('observedStateInput', '')

        if not application:
            return jsonify({'error': 'No application selected'}), 400

        instruction_descriptions = user_input.split('\n')
        instructions = [f'{i}' for i in range(0, len(instruction_descriptions))]

        results = {}
        console_output = ''
        match_plots = []

        # Update search number of episodes ONLY for instruction match
        training_episodes = data.get('trainingEpisodes', 1000)
        
        self.config.update({
            'problem_type': data.get('problemType', 'Default'),
            'number_training_episodes': int(training_episodes),
        })

        # Use default config for search agent
        self.ExperimentConfig = self.config.copy()

        engine = self.pull_app_data[application]['engine']
        local_config = self.pull_app_data[application]['local_configs'][config_input]
        adapters = self.pull_app_data[application]['adapters']
        if len(self.pull_app_data[application]['prerender_data'].keys()) > 0:
            print("Pre-rendered data found...")
            observed_states = self.pull_app_data[application]['prerender_data'][observed_states_filename]
            self.elsci_run = elsci_search(Config=self.ExperimentConfig,
                                          LocalConfig=local_config,
                                          Engine=engine, Adapters=adapters,
                                          save_dir=self.global_save_dir,
                                          number_exploration_episodes=self.num_explor_epi,
                                          match_sim_threshold=0.9,
                                          observed_states=observed_states)
        else:
            print("No pre-rendered data found...")
            self.elsci_run = elsci_search(Config=self.ExperimentConfig,
                                          LocalConfig=local_config,
                                          Engine=engine, Adapters=adapters,
                                          save_dir=self.global_save_dir,
                                          number_exploration_episodes=self.num_explor_epi,
                                          match_sim_threshold=0.9,
                                          observed_states=None)
            observed_states = self.elsci_run.search()
            with open(os.path.join(self.uploads_dir, 'observed_states.txt'), 'w') as f:
                json.dump(observed_states, f)

        best_match_dict, instruction_results = self.elsci_run.match(
            action_cap=5,
            instructions=instructions,
            instr_descriptions=instruction_descriptions
        )
        results[application] = best_match_dict
        if application not in self.instruction_results:
            self.instruction_results[application] = {}
        self.instruction_results[application]['instr_' + str(self.global_input_count)] = instruction_results

        try:
            console_output += f'<br><b>Results for {application}:</b><br>'
            for n, instr in enumerate(list(results[application].keys())):
                if results[application][instr] is None:
                    console_output += '<b>' + str(n + 1) + ' - ' + instruction_descriptions[n] + ':</b> <i>No match found</i><br>'
                else:
                    best_match = results[application][instr]['best_match']
                    console_output += '<b>' + str(n + 1) + ' - ' + instruction_descriptions[n] + ':</b> <i>' + results[application][instr]['sub_goal'] + '</i><br>'

                    plot_filename = f'match_plot_{n}.png'
                    plot_path = os.path.abspath(os.path.join(self.uploads_dir, plot_filename))
                    print(f"Saving plot to (absolute path): {plot_path}")

                    engine_dummy = engine(local_config)
                    engine_dummy.reset() # Reset required by gym environments
                    instr_match_plot = engine_dummy.render(best_match)
                    instr_match_plot_filename = f'current_state_plot_{n}.png'
                    instr_match_plot_path = os.path.abspath(os.path.join(self.uploads_dir, instr_match_plot_filename))
                    instr_match_plot.savefig(instr_match_plot_path)

                    if os.path.exists(instr_match_plot_path):
                        print(f"Current state plot created successfully at {instr_match_plot_path}")
                        print(f"File size: {os.path.getsize(instr_match_plot_path)} bytes")
                        match_plots.append(f'uploads/{instr_match_plot_filename}')
                    else:
                        print(f"Error: Current state plot not created at {instr_match_plot_path}")

        except Exception as e:
            print(f"Error in process_input: {str(e)}")
            raise

        prerender_image = self.get_prerender_image(application)

        return jsonify({
            'console_output': console_output,
            'matchPlots': match_plots,
            'prerenderImage': prerender_image
        })

    def get_prerender_image(self, application):
        image_data = self.pull_app_data[application]['prerender_images']
        image_paths = []
        if image_data:
            for image_name, image_content in image_data.items():
                image_path = os.path.join(self.uploads_dir, image_name)
                with open(image_path, 'wb') as img_file:
                    img_file.write(image_content)
                image_paths.append(f'uploads/{image_name}')
        return image_paths
        

    def train_model(self):
        data = request.json
        application = data.get('selectedApps', [])[0]
        config_input = data.get('localConfigInput', '')
        selected_plot = data.get('selectedPlot', '')
        
        if application not in self.instruction_results_validated:
            print(f"No instruction results found for {application}")
            return jsonify({'error': 'No instruction results found for the selected application'}), 400
        
        engine = self.pull_app_data[application]['engine']
        local_config = self.pull_app_data[application]['local_configs'][config_input]
        adapters = self.pull_app_data[application]['adapters']

        # Get the selected experiment config (use default if none selected)
        experimentConfigSelect = data.get('experimentConfigSelect', '')
        # 1) Selection is made, 2) Selection exists, 3) Selection is in the experiment configs
        if (experimentConfigSelect is not None)&(experimentConfigSelect != ''):
            if experimentConfigSelect in self.pull_app_data[app]['experiment_configs']:
                if self.pull_app_data[app]['experiment_configs'][experimentConfigSelect] is not None:
                    self.ExperimentConfig = self.pull_app_data[app]['experiment_configs'][experimentConfigSelect].copy()
        else:
            self.ExperimentConfig = self.config.copy()

        # --- Update Experiment Parameters with User Input ---
        selected_agents = data.get('selectedAgents', ['Qlearntab'])
        training_repeats = data.get('trainingRepeats', 5)
        training_seeds = data.get('trainingSeeds', 1)
        test_episodes = data.get('testEpisodes', 200)
        test_repeats = data.get('testRepeats', 10)
        alpha = data.get('alpha', 0.1)
        gamma = data.get('gamma', 0.95)
        epsilon = data.get('epsilon', 0.2)
        epsilon_step = data.get('epsilonStep', 0.01)

        self.ExperimentConfig.update({
            'problem_type': data.get('problemType', 'Default'),
            'number_training_repeats': int(training_repeats),
            'number_training_seeds': int(training_seeds),
            'number_test_episodes': int(test_episodes),
            'number_test_repeats': int(test_repeats),
            'agent_select': selected_agents,
            'agent_parameters': {}
        })

        # Always include Qlearntab for search agent
        self.ExperimentConfig['agent_parameters']['Qlearntab'] = {
            'alpha': float(alpha),
            'gamma': float(gamma),
            'epsilon': float(epsilon),
            'epsilon_step': float(epsilon_step)
        }

        if 'SB3_DQN' in selected_agents:
            sb_dqn_params = data.get('sbDqnParams', {})
            self.ExperimentConfig['agent_parameters']['SB3_DQN'] = {
                'policy': sb_dqn_params.get('policy', 'MlpPolicy'),
                'learning_rate': float(sb_dqn_params.get('learning_rate', 0.0001)),
                'buffer_size': int(sb_dqn_params.get('buffer_size', 1000000))
            }

        # TODO Update all adapter inputs to dict if matching to agents
        # TODO MAKE THIS A GENERIC FUNCTION CALL IN ELSCIRL
        # --> otherwise will match all adapters to all agents
        selected_adapters = data.get('selectedAdapters', [])
        if len(selected_agents) != 0:
            agent_adapter_dict = {}
            for n, agent in enumerate(selected_agents):
                adapter_list = []
                for adapter in selected_adapters:
                    # DEFAULT ADAPTERS NEED TO BE SETUP FOR DQN INPUT
                    # Reaplce 'DQN' to match to language version
                    if agent =='DQN':
                        if ('language' in adapter.lower()) or ('lang' in adapter.lower()):
                            agent = 'DQN_language'
                            self.ExperimentConfig['agent_parameters'][agent] = self.ExperimentConfig['agent_parameters']['DQN']
                            selected_agents[n] = agent
                            adapter_list.append(adapter)    
                    else:
                        adapter_list.append(adapter)
                            
                agent_adapter_dict[agent] = adapter_list
        else:
            agent_adapter_dict = list(self.pull_app_data[application]['adapters'].keys())
        self.ExperimentConfig['adapter_input_dict'] = agent_adapter_dict
        # --- End of User Input Update ---
        # Use validated instructions for training
        instruction_results = self.instruction_results_validated[application]
        print(instruction_results.keys())
        
        if not os.path.exists(self.global_save_dir+'/'+application):
            os.mkdir(self.global_save_dir+'/'+application)  
        
        # Train for all correctly validated instructions
        flat_agent_run = False
        figures_to_display = []
        for instr_key, instr_results in instruction_results.items():
            reinforced_experiment = elsciRLOptimize(
                            Config=self.ExperimentConfig, 
                            LocalConfig=local_config, 
                            Engine=engine, Adapters=adapters,
                            save_dir=self.global_save_dir+'/'+application + '/'+instr_key, 
                            show_figures = 'No', window_size=0.1,
                            instruction_path=instr_results, predicted_path=None, 
                            instruction_episode_ratio=0.1,
                            instruction_chain=True, instruction_chain_how='exact' )
            reinforced_experiment.train()
            reinforced_experiment.test()
            reinforced_experiment.render_results()
            # Get render and add to uploads
            render_results_dir = os.path.join(self.global_save_dir, application, instr_key, 'Instr_Experiment', 'render_results')
            if os.path.exists(render_results_dir):
                for file in os.listdir(render_results_dir):
                    if file.endswith('.gif'):
                        file_path = os.path.join(render_results_dir, file)
                        new_filename = f'{instr_key}_{file}'
                        shutil.copy(file_path, os.path.join(self.uploads_dir, new_filename))
                        figures_to_display.append(f'uploads/{new_filename}')
            # Baseline flat experiment
            # - only ran first time otherwise copied from prior input
            if not flat_agent_run:
                standard_experiment = STANDARD_RL(
                        Config=self.ExperimentConfig, 
                        ProblemConfig=local_config, 
                        Engine=engine, Adapters=adapters,
                        save_dir=self.global_save_dir+'/'+application + '/'+instr_key, 
                        show_figures = 'No', window_size=0.1)
                standard_experiment.train()
                standard_experiment.test()
                standard_experiment.render_results()
                flat_agent_run = True
                # Get render and add to uploads
                render_results_dir = os.path.join(self.global_save_dir, application, instr_key, 'Standard_Experiment', 'render_results')
                if os.path.exists(render_results_dir):
                    for file in os.listdir(render_results_dir):
                        if file.endswith('.gif'):
                            file_path = os.path.join(render_results_dir, file)
                            new_filename = f'No Instr_{file}'
                            shutil.copy(file_path, os.path.join(self.uploads_dir, new_filename))
                            figures_to_display.append(f'uploads/{new_filename}')
           
        # --- RESULTS ---
        if selected_plot != '':
            analysis_class = self.pull_app_data[application]['local_analysis'][selected_plot]
            for instr_key, instr_results in instruction_results.items():
                analysis_instance = analysis_class(save_dir=self.global_save_dir+'/'+application+'/'+instr_key)
                analysis_functions = [func for func in dir(analysis_instance) if callable(getattr(analysis_instance, func)) and not func.startswith("__")]

                for i, func_name in enumerate(analysis_functions):
                    func = getattr(analysis_instance, func_name)
                    fig_dict = func()
                    for figure_names,figure in fig_dict.items():
                        if figure:
                            fig_filename = f'{application}_{func_name}_{instr_key}_{figure_names}.png'
                            fig_path = os.path.join(self.uploads_dir, fig_filename)
                            figure.savefig(fig_path)
                            figures_to_display.append(f'uploads/{fig_filename}')
                            
        evaluation_types = ['training', 'testing']
        for evaluation_type in evaluation_types:
            COMBINED_VARIANCE_ANALYSIS_GRAPH(
                results_dir=self.global_save_dir+'/'+application, 
                analysis_type=evaluation_type, 
                results_to_show='simple'
            )
            variance_plot = self.global_save_dir+'/'+application+"/variance_comparison_" + evaluation_type + ".png"
            variance_filename = f'{application}_variance_analysis_{evaluation_type}.png'
            shutil.copy(variance_plot, os.path.join(self.uploads_dir, variance_filename))
            figures_to_display.append(f'uploads/{variance_filename}')

        return jsonify({
            'figures': figures_to_display
        })
    
    def upload_file(self):
        if 'file' not in request.files:
            return 'No file part'
        file = request.files['file']
        if file.filename == '':
            return 'No selected file'
        if file:
            filename = file.filename
            file.save(os.path.join(app.config['UPLOAD_FOLDER'], filename))
            return 'File uploaded successfully'

    def new_instruction(self):
        self.global_input_count += 1
        return jsonify({'status': 'success'})

    def confirm_result(self):
        data = request.json

        is_correct = data.get('isCorrect')
        user_input = data.get('userInput')

        if is_correct is None or user_input is None:
            print("Error: Invalid data provided")
            print(f"Received data: {data}")
            return jsonify({'error': 'Invalid data provided'}), 400            
        
        print(f"Received confirmation: isCorrect={is_correct}, userInput={user_input}")
        
        # Filter instruction results if validated correct
        application = data.get('selectedApps', [])[0]
        if application not in self.instruction_results_validated:
            self.instruction_results_validated[application] = {}

        if is_correct:
            self.instruction_results_validated[application]['instr_'+str(self.global_input_count)] = self.instruction_results[application]['instr_'+str(self.global_input_count)]
            self.correct_instructions.append(user_input)
            message = "<br>Great! Training an agent with this as guidance to complete the task... <br> See the results tab once training is complete."
        else:
            self.incorrect_instructions.append(user_input)
            message = "<br>Thanks for the feedback. The model will use this to improve."

        # Instr count increased on new instr or when correct/incorrect
        self.global_input_count += 1
        
        return jsonify({
            'status': 'received',
            'message': message
        })

    def get_correct_instructions(self):
        return jsonify({
            'correctInstructions': self.correct_instructions
        })

    def get_experiment_config(self, application, config):
        if not application or not config:
            return jsonify({'error': 'Missing application or config parameter'}), 400
        
        try:
            # Get the experiment config from the pull_app_data
            experiment_config = self.pull_app_data[application]['experiment_configs'][config]
            print(experiment_config)
            return jsonify({'config': experiment_config})
        except Exception as e:
            print(f"Error getting experiment config: {str(e)}")
            return jsonify({'error': 'Failed to get experiment config'}), 500

if len(sys.argv)>1:
    if 'results' in sys.argv[1]:
        if 'output' in sys.argv[1]:
            print('Using pre-rendered data from '+sys.argv[1])
            input_save_dir= sys.argv[1]
        else:
            print('Using pre-rendered data from ./output/'+sys.argv[1])
            input_save_dir= './output/'+sys.argv[1]
        if len(sys.argv)>2:
            input_explor_epi = int(sys.argv[2])
        else:
            input_explor_epi = 1000
    else:
        input_save_dir=''
        if len(sys.argv)==2:
            input_explor_epi = int(sys.argv[1])
        else:
            input_explor_epi = 1000
else:
    input_save_dir = './output'
    input_explor_epi = 1000

WebApp = WebApp(save_dir=input_save_dir,
                num_explor_epi=input_explor_epi)

@app.route('/')
def home_route():
    WebApp.load_data()
    app.config['UPLOAD_FOLDER'] = WebApp.uploads_dir
    return WebApp.home()

@app.route('/process_input', methods=['POST'])
def process_input_route():
    response = WebApp.process_input()
    return response

@app.route('/confirm_result', methods=['POST'])
def confirm_result_route():
    return WebApp.confirm_result()

@app.route('/train_model', methods=['POST'])
def train_model_route():
    return WebApp.train_model()

@app.route('/results', methods=['POST'])
def search_route():
    save_dir = request.json.get('save_dir', '')
    return jsonify({'save_dir': save_dir})

@app.route('/upload', methods=['POST'])
def upload_file_route():
    return WebApp.upload_file()

@app.route('/uploads/<path:filename>')
def uploaded_file(filename):
    try:
        upload_folder = os.path.abspath(app.config['UPLOAD_FOLDER'])
        file_path = os.path.join(upload_folder, filename)
        if not os.path.exists(file_path):
            print(f"File not found: {file_path}")
            return f"File not found: {filename}", 404
            
        print(f"Serving file from: {file_path}")
        directory = os.path.dirname(file_path)
        filename = os.path.basename(file_path)
        return send_from_directory(directory, filename, as_attachment=False)
    except Exception as e:
        print(f"Error serving file {filename}: {str(e)}")
        return f"Error: {str(e)}", 404

@app.route('/static/<path:filename>')
def static_files(filename):
    try:
        return send_from_directory('static', filename)
    except Exception as e:
        print(f"Error serving static file {filename}: {str(e)}")
        return f"Error: {str(e)}", 404

@app.route('/get_applications')
def get_applications_route():
    return WebApp.get_applications()

@app.route('/get_observed_states', methods=['POST'])
def get_observed_states_route():
    data = request.get_json()
    selected_applications = data.get('applications', [])
    observed_states = WebApp.get_observed_states(selected_applications)
    return jsonify({
        'observedStates': observed_states
    })

@app.route('/get_local_configs', methods=['POST'])
def get_local_configs_route():
    data = request.get_json()
    selected_application = data.get('application', [])
    local_configs = WebApp.get_local_configs(selected_application)
    return jsonify({
        'localConfigs': local_configs
    })

@app.route('/get_plot_options', methods=['POST'])
def get_plot_options_route():
    data = request.get_json()
    selected_application = data.get('application', '')
    plot_options = WebApp.get_plot_options(selected_application)
    return jsonify({
        'plotOptions': plot_options
    })

@app.route('/get_all_options')
def get_all_options_route():
    return WebApp.get_all_options()

@app.route('/get_prerender_image', methods=['POST'])
def get_prerender_image_route():
    data = request.get_json()
    application = data.get('application', '')
    if not application:
        return jsonify({'error': 'No application selected'}), 400
    image_paths = WebApp.get_prerender_image(application)
    if image_paths:
        return jsonify({'imagePaths': image_paths})
    return jsonify({'error': 'No prerender images found'}), 404

@app.route('/new_instruction', methods=['POST'])
def new_instruction_route():
    response = WebApp.new_instruction()
    return response

@app.route('/get_correct_instructions')
def get_correct_instructions_route():
    return WebApp.get_correct_instructions()

@app.route('/load_data')
def load_data_route():
    WebApp.global_save_dir = ''
    WebApp.load_data()
    return jsonify({'status': 'success'})

@app.route('/get_adapters', methods=['POST'])
def get_adapters_route():
    data = request.get_json()
    selected_application = data.get('application', '')
    adapters = WebApp.get_adapters(selected_application)
    return jsonify({
        'adapters': adapters
    })

@app.route('/get_experiment_config', methods=['POST'])
def get_experiment_config_route():
    data = request.get_json()
    application = data.get('application', '')
    config = data.get('config', '')
    
    if not application or not config:
        return jsonify({'error': 'Missing application or config parameter'}), 400
        
    try:
        # Get the experiment config from the pull_app_data
        experiment_config = WebApp.pull_app_data[application]['experiment_configs'][config]
        return jsonify({'config': experiment_config})
    except Exception as e:
        print(f"Error getting experiment config: {str(e)}")
        return jsonify({'error': 'Failed to get experiment config'}), 500

if __name__ == '__main__':
    if not os.path.exists(os.path.join(WebApp.global_save_dir, 'uploads')):
        os.makedirs(os.path.join(WebApp.global_save_dir, 'uploads'))
    app.run(debug=True)